// vincia01: 
// This is a simple test program to run Z decays at LEP I
// interfaced to ROOT, including a runtime updated display
// Copyright P. Skands (2010)
// (Adapted from: main09.cc, Copyright T. Sjostrand 2008)

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
// FASTJET
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia01-root.cmnd");
  // Uncomment the next line instead to run standalone Pythia for comparison:
  // VinciaPlugin vincia(&pythia,"pythia01-root.cmnd");

  // Create VinciaRoot object 
  // containing some predefined Vincia utilities for use with ROOT
  VinciaRoot vroot; // true: display ON, false: display OFF
  //vroot.initDisplay(1800,1150);
  vroot.initDisplay(1280,760);

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event& event = pythia.event;

  // Some useful pointers
  Settings* settingsPtr = vincia.settingsPtr;        // to settings database
  //VinClu* vinCluPtr = &vincia.vinClu;                // to Vincia's 3->2 clu
  //Matching* matchingPtr = &vincia.matching;          // to Vincia's matching
  //MGInterface* mgInterfacePtr = &vincia.mgInterface; // to the MG interface

  // Extract settings to be used in the main program.  
  double eCM       = settingsPtr->parm("Beams:eCM");
  int    nEvent    = settingsPtr->mode("Main:numberOfEvents");
  int    nShow     = settingsPtr->mode("Main:timesToShow");
  bool   vinciaOn  = settingsPtr->flag("Vincia");
  int    verbose   = settingsPtr->mode("Vincia:verbose");

  // Tell VinciaRoot which verbose level to use
  vroot.setVerbose(verbose);

  // Pythia tune (if using Pythia):
  int    pyTune    = settingsPtr->mode("Tune:ee");

  //************************************************************************

  // Shorthands for Vincia shower settings
  int    evType  = settingsPtr->mode("Vincia:evolutionType");
  int    kinType = settingsPtr->mode("Vincia:kineMapType");
  int    hadType = settingsPtr->mode("Vincia:cutoffType");
  double hadScale       = settingsPtr->parm("Vincia:cutoffScale");
  bool   useFiniteTerms = settingsPtr->flag("Vincia:useFiniteTerms"); 
  int    orderingMode   = settingsPtr->mode("Vincia:orderingMode");  
  bool   sectorOrdering = settingsPtr->flag("Vincia:sectorOrdering");
  string antennaFile    = settingsPtr->word("Vincia:antennaFile");
  // Pythia Hadronization Tune
  int    pyHadTune      = settingsPtr->mode("Tune:ee");
  // Matching settings
  int    matchingStrategy = settingsPtr->mode("Vincia:matchingStrategy");
  int    maxMatchTree     = settingsPtr->mode("Vincia:maxMatchTree");
  int    maxMatchLoop     = settingsPtr->mode("Vincia:maxMatchLoop");
  // Antenna type
  string antType="def";
  if (antennaFile == "antennae/Antennae-Min.cmnd")      antType="min";
  else if (antennaFile == "antennae/Antennae-Med.cmnd") antType="med";
  else if (antennaFile == "antennae/Antennae-Max.cmnd") antType="max";
  else if (antennaFile == "antennae/Antennae-Ariadne.cmnd") antType="ari";
  else if (sectorOrdering) antType="sct";
  if (!useFiniteTerms) antType = (!sectorOrdering) ? "s-glb" : "s-sct";

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythia.init( 11, -11, eCM);

  // List changed settings
  pythia.settings.listChanged();

  // Check that Z0 decay channels set correctly.
  pythia.particleData.listChanged();
 
  //***************************************************************************
  // FASTJET
  //***************************************************************************

  // Fastjet analysis - select algorithm and parameters
  fastjet::Strategy               strategy = fastjet::Best;
  fastjet::RecombinationScheme    recombScheme = fastjet::E_scheme;
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::ee_kt_algorithm,
                                      recombScheme, strategy);
  // Fastjet input
  std::vector <fastjet::PseudoJet> fjInputs;

  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  outfile<<"vc01";
  if (not vinciaOn) {
    outfile<<"-pyt";
    outfile<<"-tun"<<pyHadTune;
    outfile<<"-q"<<settingsPtr->mode("TimeShower:nGluonToQuark");      
  } else {
    outfile<<"-"<<antType;
    outfile<<"-ev"<<evType;
    outfile<<"-kin"<<kinType;
    outfile<<"-had"<<hadType;
    double hadScaleRounded = int(hadScale*10+0.5)/10.;
    outfile<<"q"<<setprecision(2)<<hadScaleRounded;
    outfile<<"tun"<<pyHadTune;
    if (orderingMode == 1) outfile<<"-ord";
    if (orderingMode == 0) outfile<<"-uno";
    if (orderingMode == 2) outfile<<"-uqe";
    if (orderingMode == 3) outfile<<"-upt";

    if (settingsPtr->flag("Vincia:matching")) {
      outfile<<"-m"<<matchingStrategy;
      if (settingsPtr->mode("Vincia:maxMatchLoop") >= 1) { 
	
	outfile<<"-L"<<setprecision(2)<<int(maxMatchLoop*10+0.5)/10.;
      } 
      outfile<<"-T"<<int(maxMatchTree*10+0.5)/10.;;
    }
    //    if (settingsPtr->flag("Vincia:NLLmatching")) outfile<<"-nll";
    outfile<<"-nq"<<settingsPtr->mode("Vincia:nGluonToQuark");      
  }
  if (settingsPtr->flag("HadronLevel:all")) outfile<<"-had";
  outfile<<".root";

  //***************************************************************************
  // Book Histograms

  // Book alphaS histogram
  TH1F *alphaS=new TH1F("alphaS","alpha_S vs Q",200,0,200.0);
  // Add to Vincia's list for run-time display and post-run saving
  // logy = false, normalize=false, showStatErr=false
  //  vroot.addTH1FPtr(alphaS,false,false,false);
  // Set properties 
  alphaS->SetFillColor(3); 
  alphaS->SetMinimum(0.0);
  alphaS->SetMaximum(0.5);
  // Fill histogram
  for (double Q=0.5;Q<200.0;Q += 1.) {
    double as   = vincia.shower.alphaS.alphaS(pow2(Q));
    if (as < 10.0) alphaS->Fill(Q,as);
  }

  // Event Weights
  TH1F *HnPN=new TH1F("Weights","Event Weights", 100, -2.5, 2.5);
  //  vroot.addTH1FPtr(HnPN,true,false,false);
  HnPN->SetFillColor(3); 

  // Quark Pairs
  TH1F *HnQuarks=new TH1F("nQuarks","Number of Quark Pairs", 15, -0.5, 14.5);
  //  vroot.addTH1FPtr(HnQuarks,true,true,true);
  HnQuarks->SetFillColor(3); 
  HnQuarks->SetMinimum(0.0001);
  HnQuarks->SetMaximum(0.75);

  // nPartons
  TH1F *HnPartons=new TH1F("nPartons","Number of Partons", 60, -0.5, 59.5);
  //  vroot.addTH1FPtr(HnPartons,true,true,true);
  HnPartons->SetFillColor(3); 
  HnPartons->SetMinimum(0.001);
  HnPartons->SetMaximum(0.25);

  // Thrust
  TH1F *HThrust=new TH1F("Thrust","1-T (udsc)", 50, 0.0, 0.4);
  double wThrust = 50.0/0.4;
  vroot.addTH1FPtr(HThrust,true,true,true,"data/L3-91-Thrust-udsc.vec");
  HThrust->SetFillColor(3); 
  HThrust->SetMinimum(0.001);
  HThrust->SetMaximum(100.);

 // Thrust Major
  TH1F *HTmaj=new TH1F("Tmaj","Thrust Major", 100, 0.0, 0.7);
  double wTmaj = 100.0/0.7;
  vroot.addTH1FPtr(HTmaj,true,true,true,"data/LEP91-Major.vec");
  HTmaj->SetFillColor(3); 
  HTmaj->SetMinimum(0.001);
  HTmaj->SetMaximum(20.);

  // Thrust Minor
  TH1F *HTmin=new TH1F("Tmin","Thrust Minor", 50, 0.0, 0.4);
  double wTmin = 50.0/0.4;
  vroot.addTH1FPtr(HTmin,true,true,true,"data/Delphi-91-Minor.vec");
  HTmin->SetFillColor(3); 
  HTmin->SetMinimum(0.001);
  HTmin->SetMaximum(20.);
 
  // Oblateness
  TH1F *HObl=new TH1F("Oblateness","Oblateness",100, 0.0, 0.6);
  double wObl = 100.0/0.6;
  vroot.addTH1FPtr(HObl,true,true,true,"data/LEP91-Oblateness.vec");
  HObl->SetFillColor(3); 
  HObl->SetMinimum(0.001);
  HObl->SetMaximum(20.);

  // C
  TH1F *HC=new TH1F("C","C (udsc)", 60, 0.0, 1.0);
  double wC = 60.0/1.0;
  vroot.addTH1FPtr(HC,true,true,true,"data/L3-91-C-udsc.vec");
  HC->SetFillColor(3); 
  HC->SetMinimum(0.001);
  HC->SetMaximum(100.);

  // D
  TH1F *HD=new TH1F("D","D (udsc)", 50, 0.0, 0.8);
  double wD = 50.0/0.8;
  vroot.addTH1FPtr(HD,true,true,true,"data/L3-91-D-udsc.vec");
  HD->SetFillColor(3); 
  HD->SetMinimum(0.001);
  HD->SetMaximum(100.);
  
  // PTIN
  TH1F *HPtIn=new TH1F("PtIn","PTIN (charged)",100, 0.0, 15.0);
  double wPtIn = 100.0/15.0;
  vroot.addTH1FPtr(HPtIn,true,true,true,"data/Delphi-91-PTIN.vec");
  HPtIn->SetFillColor(3); 
  HPtIn->SetMaximum(10.01); 
  HPtIn->SetMinimum(0.00001);   

  // PTOUT
  TH1F *HPtOut=new TH1F("PtOut","PTOUT (charged)",100, 0.0, 5.0);
  double wPtOut = 100.0/5.0;
  vroot.addTH1FPtr(HPtOut,true,true,true,"data/Delphi-91-PTOUT.vec");
  HPtOut->SetFillColor(3); 
  HPtOut->SetMaximum(10.01); 
  HPtOut->SetMinimum(0.00001);   

  // Jet Rates
  TH1F *HLnY12=new TH1F("LnY12","Ln(1/y12)",100, 0.0, 1.3);
  double wY12 = 100.0/1.3;
  vroot.addTH1FPtr(HLnY12,true,true,true,"data/LEP91-LnY12.vec");
  HLnY12->SetFillColor(3); 
  HLnY12->SetMaximum(10.1); 
  HLnY12->SetMinimum(0.00001); 

  // Jet Rates
  TH1F *HLnY23=new TH1F("LnY23","Ln(1/y23)",100, 1.0, 11.0);
  double wY23 = 100.0/10.;
  vroot.addTH1FPtr(HLnY23,true,true,true,"data/LEP91-LnY23.vec");
  HLnY23->SetFillColor(3); 
  HLnY23->SetMaximum(1.01); 
  HLnY23->SetMinimum(0.00001); 

  // Jet Rates
  TH1F *HLnY34=new TH1F("LnY34","Ln(1/y34)",100, 2.0, 11.0);
  double wY34 = 100.0/9.;
  vroot.addTH1FPtr(HLnY34,true,true,true,"data/LEP91-LnY34.vec");
  HLnY34->SetFillColor(3); 
  HLnY34->SetMaximum(1.01); 
  HLnY34->SetMinimum(0.0001); 

  // Jet Rates
  TH1F *HLnY45=new TH1F("LnY45","Ln(1/y45)",100, 3.0, 12.0);
  double wY45 = 100.0/9.0;
  vroot.addTH1FPtr(HLnY45,true,true,true,"data/LEP91-LnY45.vec");
  HLnY45->SetFillColor(3); 
  HLnY45->SetMaximum(1.01); 
  HLnY45->SetMinimum(0.0001); 

  // Jet Rates
  TH1F *HLnY56=new TH1F("LnY56","Ln(1/y56)",100, 4.0, 13.0);
  double wY56 = 100.0/9.;
  vroot.addTH1FPtr(HLnY56,true,true,true,"data/LEP91-LnY56.vec");
  HLnY56->SetFillColor(3); 
  HLnY56->SetMaximum(1.01); 
  HLnY56->SetMinimum(0.0001); 

  // Heavy-Jet Mass (3-jet obs)
  TH1F *HMjH=new TH1F("HeavyJetMass","Heavy Jet Mass",100, 0.0, 0.5);
  double wMjH = 100.0/0.5;
  vroot.addTH1FPtr(HMjH,true,true,true,"data/Delphi-91-MH.vec");
  HMjH->SetFillColor(3); 
  HMjH->SetMaximum(100.01); 
  HMjH->SetMinimum(0.001); 
  
  // Light-Jet Mass (4-jet obs)
  TH1F *HMjL=new TH1F("LightJetMass","Light Jet Mass",100, 0.0, 0.2);
  double wMjL = 100.0/0.2;
  vroot.addTH1FPtr(HMjL,true,true,true,"data/Delphi-91-ML.vec");
  HMjL->SetFillColor(3); 
  HMjL->SetMaximum(100.01); 
  HMjL->SetMinimum(0.001);   
  
  // Wide Jet Broadening
  TH1F *HBjWide=new TH1F("WideJetBroadening","Wide Jet Broadening",100, 0.0, 0.3);
  double wBjWide = 100.0/0.3;
  vroot.addTH1FPtr(HBjWide,true,true,true,"data/Delphi-91-BWide.vec");
  HBjWide->SetFillColor(3); 
  HBjWide->SetMaximum(100.01); 
  HBjWide->SetMinimum(0.01);   

  // Total Jet Broadening
  TH1F *HBjTot=new TH1F("TotalJetBroadening","Total Jet Broadening",100, 0.0, 0.4);
  double wBjTot = 100.0/0.4;
  vroot.addTH1FPtr(HBjTot,true,true,true,"data/Delphi-91-BTot.vec");
  HBjTot->SetFillColor(3); 
  HBjTot->SetMaximum(10.01); 
  HBjTot->SetMinimum(0.01);   

  // Jet Shape
  TH1F *HSj=new TH1F("JetShape","Jet Shape",100, 0.0, 1.0);
  double wSj = 100.0/1.0;
  vroot.addTH1FPtr(HSj,true,true,true);
  HSj->SetFillColor(3); 
  HSj->SetMaximum(1.01); 
  HSj->SetMinimum(0.0001);   

  // Ln(1/x)
  TH1F *Hx=new TH1F("LnXInv","Ln(1/x) (udsc)", 100, 0.0, 8.0);
  double wx = 100.0/8.0;
  vroot.addTH1FPtr(Hx,true,true,true,"data/L3-91-LogInvx-udsc.vec");
  Hx->SetFillColor(3); 
  Hx->SetMinimum(0.0001);
  Hx->SetMaximum(1.);

  // nCharged
  TH1F *HnCharged=new TH1F("nCharged","Number of Charged Particles (udsc)", 60, -0.5, 59.5);
  vroot.addTH1FPtr(HnCharged,true,true,true,"data/L3-91-Nch-udsc.vec");
  HnCharged->SetFillColor(3); 
  HnCharged->SetMinimum(0.00001);
  HnCharged->SetMaximum(0.5);

  // Mesons
  TH1F *HMesons=new TH1F("Mesons","nMesons/nCharged",10, -0.5, 9.5);
  HMesons->GetXaxis()->SetBinLabel(1,"#pi^{#pm}");
  HMesons->GetXaxis()->SetBinLabel(2,"#pi^{0}");
  HMesons->GetXaxis()->SetBinLabel(3,"#rho^{#pm}");
  HMesons->GetXaxis()->SetBinLabel(4,"#rho^{0}");
  HMesons->GetXaxis()->SetBinLabel(5,"#omega");
  HMesons->GetXaxis()->SetBinLabel(6,"#phi");
  HMesons->GetXaxis()->SetBinLabel(7,"K^{#pm}");
  HMesons->GetXaxis()->SetBinLabel(8,"K^{*#pm}");
  HMesons->GetXaxis()->SetBinLabel(9,"#eta");
  HMesons->GetXaxis()->SetBinLabel(10,"#eta'");
  vroot.addTH1FPtr(HMesons,true,false,false,"data/LEP91-mesons.vec");
  HMesons->SetFillColor(3); 
  HMesons->SetMaximum(1.01); 
  HMesons->SetMinimum(0.001); 

  // Baryons
  TH1F *HBaryons=new TH1F("Baryons","nBaryons/nCharged",9, -0.5, 8.5);
  HBaryons->GetXaxis()->SetBinLabel(1,"p");
  HBaryons->GetXaxis()->SetBinLabel(2,"#Lambda");
  HBaryons->GetXaxis()->SetBinLabel(3,"#Sigma^{#pm}");
  HBaryons->GetXaxis()->SetBinLabel(4,"#Sigma^{0}");
  HBaryons->GetXaxis()->SetBinLabel(5,"#Delta");
  HBaryons->GetXaxis()->SetBinLabel(6,"#Sigma^{*}");
  HBaryons->GetXaxis()->SetBinLabel(7,"#Xi^{#pm}");
  HBaryons->GetXaxis()->SetBinLabel(8,"#Xi^{*0}");
  HBaryons->GetXaxis()->SetBinLabel(9,"#Omega");
   vroot.addTH1FPtr(HBaryons,true,false,false,"data/LEP91-baryons.vec");
  HBaryons->SetFillColor(3); 
  HBaryons->SetMaximum(0.101); 
  HBaryons->SetMinimum(0.00001); 

  // Heavy Mesons 
  TH1F *HHeavyMesons=new TH1F("HeavyMesons","nHeavyMesons/nCharged",9, -0.5, 8.5);
  vroot.addTH1FPtr(HHeavyMesons,true,false,false,"data/LEP91-heavymesons.vec");
  HHeavyMesons->SetFillColor(3); 
  HHeavyMesons->SetMaximum(0.101); 
  HHeavyMesons->SetMinimum(0.00001); 

  // USER HISTOGRAM
  // The example below fills it with the event weights (normally unity)
  TH1F *HUser=new TH1F("User","USER",41, -2.05, 2.05);
  vroot.addTH1FPtr(HUser,true,false,false);
  HUser->SetFillColor(2); 

  // Event analysis tools
  Thrust Thrust(1);
  Sphericity SphLin(1,2);
  ClusterJet ktClu("Durham",2,2);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // Counter for negative-weight events
  double nNegative = 0.0;
  double nPositive = 0.0;
  double evwt=1.0;
  bool firstEvent = true;

  // Number of quarks, partons, and charged particles
  double nQsum=0.0;
  double nPsum=0.0;
  double nCsum=0.0;
  double nMesSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nBarSum[10] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nHMesSum[10] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {
    if (iEvent%(max(1,nEvent/max(nShow,1))) == 0)  
      cout << " Now begin event " << iEvent << endl;

    if (!pythia.next()) continue;

    // Check for negative weights
    if (vinciaOn) evwt = vincia.weight();
    if (evwt < 0) nNegative += abs(evwt); 
    else nPositive += evwt;
    HnPN->Fill(evwt,1.0);
    HUser->Fill(evwt,1.0);

    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;
    bool isB = false;
    // Reset Fastjet input
    fjInputs.resize(0);

    for (int i=1;i<event.size();i++) {
      if (abs(event[i].id()) == 5) isB = true;
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) {
	nc++;
	double LnXInv=log(double(1.0/(2.0*event[i].pAbs()/eCM)));
	if (! isB) Hx->Fill(LnXInv,evwt*wx);
      }
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
	// Count up partons and quarks
	if (event[i].isQuark() || event[i].isGluon() ) np++;
	if (event[i].isQuark()) nq++;	
      }
      // Find identified particles
      /* HMesons:
	# 211 0: e+ e- --> pi+- X
	# 111 1: e+ e- --> pi0 X
	# 213 2: e+ e- --> rho(770)+- X
	# 113 3: e+ e- --> rho(770)0 X
	# 223 4: e+ e- --> omega(782) X
	# 333 5: e+ e- --> phi(1020) X
	# 321 6: e+ e- --> K+- X
	# 323 7: e+ e- --> K*(892)+- X
	# 221 8: e+ e- --> eta X
	# 331 9: e+ e- --> etaprime(958) X
	# 10111 10: e+ e- --> a0(980) X
	# 10221 11: e+ e- --> f0(980) X
      */
      /* HBaryons:
	 # 2212      0 : e+ e- --> p (pbar) X
	 # 3122      1: e+ e- --> Lambda (Lambdabar) X
	 # 3222+3112 2: e+ e- --> Sigma+- (Sigmabar+-) X
	 # 3212      3: e+ e- --> Sigma0 (Sigmabar0) X
	 # 2224      4 : e+ e- --> Delta(1232)++ (Delta(1232)--) X
	 # 3224+3114 5: e+ e- --> Sigma(1385)+- (Sigmabar(1385)+-) X
	 # 3312      6: e+ e- --> Xi- (Xibar+) X
	 # 3324      7: e+ e- --> Xi(1530)0 (Xibar(1530)0) X
	 # 3334      8: e+ e- --> Omega- (Omegabar+) X 
      */
      /* HHMesons:
	 # 411    0:  e+ e- --> D+- X
         # 421    1: e+ e- --> D0 (Dbar0) X
         # 413    2: e+ e- --> D*(2010)+- X
         # 443    3: e+ e- --> J/psi X
         # 20443  4: e+ e- --> chi(c1) X
         # 100443 5: e+ e- --> psi(3685) X  	
      */
      int idp = abs(event[i].id());
      if (idp == 211) nMesSum[0] += evwt;
      if (idp == 111) nMesSum[1]+= evwt;
      if (idp == 213) nMesSum[2]+= evwt;
      if (idp == 113) nMesSum[3]+= evwt;
      if (idp == 223) nMesSum[4]+= evwt;
      if (idp == 333) nMesSum[5]+= evwt;
      if (idp == 321) nMesSum[6]+= evwt;
      if (idp == 323) nMesSum[7]+= evwt;
      if (idp == 221) nMesSum[8]+= evwt;
      if (idp == 331) nMesSum[9]+= evwt;
      if (idp == 10111) nMesSum[10]+= evwt;
      if (idp == 10221) nMesSum[11]+= evwt;
      if (idp == 2212) nBarSum[0] += evwt;
      if (idp == 3122) nBarSum[1] += evwt;
      if (idp == 3222) nBarSum[2] += evwt; 
      if (idp == 3112) nBarSum[2] += evwt; 
      if (idp == 3212) nBarSum[3] += evwt; 
      if (idp == 2224) nBarSum[4] += evwt; 
      if (idp == 3114) nBarSum[5] += evwt; 
      if (idp == 3224) nBarSum[5] += evwt; 
      if (idp == 3312) nBarSum[6] += evwt; 
      if (idp == 3324) nBarSum[7] += evwt; 
      if (idp == 3334) nBarSum[8] += evwt; 
      if (idp == 411) nHMesSum[0] += evwt;
      if (idp == 421) nHMesSum[1] += evwt;
      if (idp == 413) nHMesSum[2] += evwt;
      if (idp == 443) nHMesSum[3] += evwt;
      if (idp == 20443) nHMesSum[4] += evwt;
      if (idp == 100443) nHMesSum[5] += evwt;

      // Now decide what to pass to FastJet

      // Final state only
      if (!pythia.event[i].isFinal())        continue;
      
      // No neutrinos
      if (pythia.event[i].idAbs() == 12 || pythia.event[i].idAbs() == 14 ||
          pythia.event[i].idAbs() == 16)     continue;
      
      // Store as input to Fastjet
      fjInputs.push_back( fastjet::PseudoJet (pythia.event[i].px(),
					      pythia.event[i].py(), 
					      pythia.event[i].pz(),
					      pythia.event[i].e() ) );
    }

    if (fjInputs.size() == 0) {
      cout << "Error: event with no final state particles" << endl;
      continue;
    }

    // Run Fastjet algorithm
    fastjet::ClusterSequence clustSeq(fjInputs, *jetDef);

    // For the first event, print the FastJet details
    if (firstEvent) {
      cout << "Ran " << jetDef->description() << endl;
      cout << "Strategy adopted by FastJet was "
           << clustSeq.strategy_string() << endl << endl;
      firstEvent = false;
      
    }

    // Extract merging scales
    double lny56FJ = log(clustSeq.exclusive_ymerge(5));
    double lny45FJ = log(clustSeq.exclusive_ymerge(4));
    double lny34FJ = log(clustSeq.exclusive_ymerge(3));
    double lny23FJ = log(clustSeq.exclusive_ymerge(2));
    double lny12FJ = log(clustSeq.exclusive_ymerge(1));


    // Fill Histograms
    HnQuarks->Fill(0.5*nq,evwt);
    HnPartons->Fill(1.0*np,evwt);    
    if (! isB) HnCharged->Fill(1.0*nc,evwt);
    nQsum += 0.5*nq;
    nPsum += np;
    nCsum += nc;

    // Fill PID histograms
    HMesons->Reset();
    for (int iMes=0; iMes<=11; iMes++) {
      HMesons->Fill(double(iMes),nMesSum[iMes]/nCsum);
    }
    HBaryons->Reset();
    for (int iBar=0; iBar<=8; iBar++) {
      HBaryons->Fill(double(iBar),nBarSum[iBar]/nCsum);
    }
    HHeavyMesons->Reset();
    for (int iHMes=0; iHMes<=8; iHMes++) {
      HHeavyMesons->Fill(double(iHMes),nHMesSum[iHMes]/nCsum);
    }
	
    // Event Shape analysis
    Thrust.analyze( event );
    if (! isB) HThrust->Fill(1-Thrust.thrust(),wThrust*evwt);
    HObl->Fill(Thrust.oblateness(),wObl*evwt);
    HTmaj->Fill(Thrust.tMajor(),wTmaj*evwt);
    HTmin->Fill(Thrust.tMinor(),wTmin*evwt);    

    // Histogram jet masses
    Vec4 thrustAxis = Thrust.eventAxis(1);   
    Vec4 majorAxis = Thrust.eventAxis(2);   
    Vec4 minorAxis = Thrust.eventAxis(3);   
    Vec4 pPlus, pMinus;
    double bPlus(0.), bMinus(0.), pPlusAbs(0.), pMinusAbs(0.), pTout(0.), pTin(0.);
    // Loop over event to compute masses
    for (int i=0; i<event.size(); i++) {
      if (not event[i].isFinal() ) continue;
      int id = event[i].id();
      if (abs(id) == 12 || abs(id) == 14 || abs(id) == 16) continue;
      double dotp = event[i].px()*thrustAxis.px()
	+ event[i].py()*thrustAxis.py()
	+ event[i].pz()*thrustAxis.pz();
      double pAbs = event[i].pAbs();
      double cosT = dotp/pAbs/thrustAxis.pAbs();
      double sinT = sqrt(1.-pow2(cosT));
      if (dotp > 0) {
	pPlus     += event[i].p();
	bPlus     += abs(pAbs * sinT);
	pPlusAbs  += pAbs;
      } else {
	pMinus += event[i].p();
	bMinus    += abs(pAbs * sinT);
	pMinusAbs += pAbs;
      }
      if (event[i].isCharged()) {
	pTin  = abs(event[i].px()*majorAxis.px() + event[i].py()*majorAxis.py()
		    +event[i].pz()*majorAxis.pz());
	pTout = abs(event[i].px()*minorAxis.px() + event[i].py()*minorAxis.py()
		    +event[i].pz()*minorAxis.pz());
	HPtIn->Fill(pTin,wPtIn*evwt);
	HPtOut->Fill(pTout,wPtOut*evwt);
      }
    }
    bPlus  /= 2.*(pPlusAbs+pMinusAbs);
    bMinus /= 2.*(pMinusAbs+pPlusAbs);    
    double mPlus  = pPlus.mCalc();
    double mMinus = pMinus.mCalc();
    HMjH->Fill(pow2(max(mPlus,mMinus)/eCM),wMjH*evwt);
    HMjL->Fill(pow2(min(mPlus,mMinus)/eCM),wMjL*evwt);
    HBjWide->Fill(max(bPlus,bMinus),wBjWide*evwt);
    HBjTot->Fill(bPlus+bMinus,wBjTot*evwt);   

    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
		    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
		    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
	*SphLin.eigenValue(3);
      if (! isB) HC->Fill(evC,wC*evwt);
      if (! isB) HD->Fill(evD,wD*evwt);
    }

    // Jet rate analysis
    // Start by always clustering to exactly one jet
    /*
    ktClu.analyze(event,1.0,0.0,1,1);
    double lny12=log(ktClu.distance(0)/eCM/eCM);
    double lny23=log(ktClu.distance(1)/eCM/eCM);
    double lny34=log(ktClu.distance(2)/eCM/eCM);
    double lny45=log(ktClu.distance(3)/eCM/eCM);
    double lny56=log(ktClu.distance(4)/eCM/eCM);
    */

    HLnY12->Fill(-lny12FJ,wY12*evwt);
    HLnY23->Fill(-lny23FJ,wY23*evwt);
    HLnY34->Fill(-lny34FJ,wY34*evwt);
    HLnY45->Fill(-lny45FJ,wY45*evwt);
    HLnY56->Fill(-lny56FJ,wY56*evwt);
 
    // Periodically update displays
    if (iEvent == 9 || iEvent%500 == 99|| iEvent == nEvent-1) {
      vroot.updateDisplay();       
      if ( (iEvent >= 1000 && iEvent%5000==99) || iEvent == nEvent-1) {
	string plotName="lastDisplay.gif";
	if (not vinciaOn) {
	  plotName="PythiaDefault.gif";
	  if (pyTune == 1) plotName="PythiaTune1.gif";
	  if (pyTune == 2) plotName="PythiaTune2.gif";
	  if (pyTune == 3) plotName="PythiaTune3.gif";
	}
	vroot.saveDisplay(plotName,"Landscape");
      }
    }

  }

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.statistics();
  if (vinciaOn) vincia.printInfo();

  // Close ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  // Open ROOT file
  TFile f(outfile.str().c_str(),"recreate");
  // Opened root file:
  cout <<"vincia01kine : opened ROOT file "<<outfile.str()<<endl;  
  alphaS->Write();
  HnPN->Write();
  HnQuarks->Write();
  HnPartons->Write();
  HnCharged->Write();
  HMesons->Write();
  HBaryons->Write();
  HHeavyMesons->Write();
  Hx->Write();
  HLnY12->Write();
  HLnY23->Write();
  HLnY34->Write();
  HLnY45->Write();
  HLnY56->Write();
  HThrust->Write();
  HTmaj->Write();
  HTmin->Write();
  HObl->Write();
  HC->Write();
  HD->Write();
  HMjH->Write();
  HMjL->Write();
  HBjWide->Write();
  HBjTot->Write();
  HPtIn->Write();
  HPtOut->Write();
  f.Close();

  cout << endl << "Enter q to terminate..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
